# this file is for documenting environment variables and has no functional value

"""
:DOMINO_AI_SYSTEM_CONFIG_PATH: For configuring the location of the ai_system_config.yaml file.
    If not set, defaults to './ai_system_config.yaml'.
:type: str

:DOMINO_AI_SYSTEM_MODEL_ID: The ID of the production AI System
:type: str
"""
